;; paper_rom - Copyright 2013 by Michael Kohn
;; Email: mike to mikekohn.net
;;   Web: http://www.mikekohn.net/
;;
;; Read data from a disk of paper with black marks of different thickness.

.avr8
.include "tn2313def.inc"
;.device ATtiny2313

; 20MHz clock
; note: CLKSEL 1111
; lowfuse = 0x8f

; disk radius = 8.0cm
; track 0 radius = 7.5cm   circumference = 47.10cm
; track 1 radius = 6.0cm   circumference = 37.68cm
; track 2 radius = 4.5cm   circumference = 28.26cm
; track 3 radius = 3.0cm   circumference = 18.84cm

; read_sensor() is rouround 0.195us

; current markers = 1.4cm, 0.7cm, 0.35cm
; track 0  @1 RPS = 30ms, 15ms, 7ms         [ 154  77 36 ]
; track 1  @1 RPS = 37ms, 19ms, 9ms         [ 190  97 46 ]
; track 2  @1 RPS = 50ms, 25ms, 12ms        [ 256 128 62 ]
; track 3  @1 RPS = 74ms, 37ms, 19ms        [ 379 190 97 ]

; Timer 0: 20,000,000 / 256 = 78,145 interrupts = 1 second
; Timer 1: 16 bit counter, counting pulses on T1

; r0  = 0
; r1  = 1
; r2  =
; r3  =
; r4  = marker length counte low
; r5  = marker length counte high
; r7  = store SREG in interrupts
; r8  = 200
; r9  = mutilplier for r8
; r10 = counter in timer1 interrupt
; r11 = servo0 value (from 10 to 20.. 14 being no movement)
; r12 = serve1 value (from 10 to 20.. 14 being center)
; r13 = temp for sending hex
; r14 = temp for sending hex
; r15 = 5
; r16 = temp counter
; r17 = temp
; r18 = temp in funct
; r19 = send_char param
; r20 = read from rs232
; r21 = counter in timer0 interrupt
; r24 = longest mark low
; r25 = longest mark high
; r26 = length of one low
; r27 = length of one high
; r30 =
; r31 = current bit count

LONG_LOW equ SRAM_START
LONG_HIGH equ SRAM_START+1
ONE_LOW equ SRAM_START+2
ONE_HIGH equ SRAM_START+3
ZERO_LOW equ SRAM_START+4
ZERO_HIGH equ SRAM_START+5

.org 0x000
  rjmp start
.org 0x4
  rjmp timer1_interrupt
.org 0xd
  rjmp timer0_interrupt

start:
  ;; I'm busy.  Don't interrupt me!
  cli

  ;; Set up stack ptr
  ;ldi r17, RAMEND>>8
  ;out SPH, r17
  ldi r17, RAMEND&255
  out SPL, r17

  ;; r0 = 0, r1 = 1, r15 = 5, r8 = 200
  clr r0
  clr r1
  inc r1
  ldi r17, 200
  mov r8, r17
  ldi r17, 5
  mov r15, r17

  ; init variables
  mov r11, r0                  ; servo0 doesn't spin
  ldi r17, 13
  mov r12, r17                 ; servo1 is centered

  ;; Set up rs232 baud rate
  ldi r17, ((20000000/(16*9600))-1) >> 8
  out UBRRH, r17
  ldi r17, ((20000000/(16*9600))-1) & 0xff
  out UBRRL, r17

  ;; Set up rs232 options
  ldi r17, (1<<UCSZ0)|(1<<UCSZ1)      ; sets up data as 8N1
  out UCSRC, r17
  ldi r17, (1<<TXEN)|(1<<RXEN)        ; enables send/receive
  out UCSRB, r17
  out UCSRA, r0

  ;; Set up PORTB
  ldi r17, 0x0d
  out DDRB, r17
  out PORTB, r1
  sbi PORTB, 7

  ;; Set up TIMER0
  ldi r17, 99                    ; compare to 100 clocks (5us @20MHz)
  out OCR0A, r17
  ldi r17, (1<<WGM01)
  out TCCR0A, r17                ; normal counting (OC0A is top, count up)
  ldi r17, (1<<CS00)             ; No prescaling
  out TCCR0B, r17

  ; Set up TIMER1
  ldi r17, (1999>>8)
  out OCR1AH, r17
  ldi r17, (1999&0xff)            ; compare to 1000 clocks (0.0001s @20MHz)
  out OCR1AL, r17

  out TCCR1C, r0
  out TCCR1A, r0
  ldi r17, (1<<CS10)|(1<<WGM12)
  out TCCR1B, r17                ; No prescaler / OCR1A is top

  ;; Set up interrupts for both timer modules
  ldi r17, (1<<OCIE0A)|(1<<OCIE1A)
  out TIMSK, r17                 ; Interrupt on Timer 0 overflow only

  ; Fine, I can be interrupted now
  sei

  ;; DEBUG
  ldi r19, 'A'
  rcall send_hex
  ;; DEBUG

main:
  sbic PINB, 7
  rjmp main

  ;; Button has been pressed, time to read data from disk
  ldi r17, 20
  mov r11, r17                 ; servo0 spins

  ldi r17, 13
  mov r12, r17                 ; servo1 is on track 0
  rcall read_track

  ldi r17, 14
  mov r12, r17                 ; servo1 is on track 1
  rcall read_track

  ;; SLOW IT?
  ;ldi r17, 21
  ;mov r11, r17                 ; servo0 spins
  ;;

  ldi r17, 15
  mov r12, r17                 ; servo1 is on track 2
  rcall read_track

  ldi r17, 16
  mov r12, r17                 ; servo1 is on track 3
  rcall read_track

  ldi r17, 13
  mov r12, r17                 ; servo1 is on track 0 (home)
  mov r11, r0                  ; servo0 doesn't spin

  rjmp main

read_track:
  ;; Give a delay (approx 50ms) for servos to move
  ldi r17, 0
pause_50ms:
  rcall read_sensor
  dec r17
  brne pause_50ms

  ;; Calibrate: find length of longest mark by spinning until 134 marks
  ;; have been read and record the longest.
  clr r24
  clr r25
  ldi r16, 134
read_another_mark:

  rcall wait_for_black
  rcall count_black

  ;; 16 bit compare to see if we should set r25:r24
  cp r5, r25
  breq check_lower
  brlo not_longer
  brsh set_longer
check_lower:
  cp r4, r24
  brlo not_longer
set_longer:
  movw r24, r4

not_longer:
  dec r16
  brne read_another_mark

  ;; DEBUG
  ;mov r19, r25
  ;rcall send_hex
  ;mov r19, r24
  ;rcall send_hex
  ;; DEBUG

  ;; Save of lengths to RAM.  Subtract some from each to take into account
  ;; that the length of each mark won't be exact every time
  movw r4, r24
  subi r24, 10
  brcc long_no_borrow
  dec r25
long_no_borrow:
  sts LONG_LOW, r24            ; NOTE: this is probably faster to use a pointer
  sts LONG_HIGH, r25           ; like X+

  ;mov r19, r25
  ;rcall send_hex
  ;mov r19, r24
  ;rcall send_hex

.if 0
  movw r24, r4                 ; ONE = LONG / 2 - 15 
  lsr r25
  ror r24
  subi r24, 15 
  brcc one_no_borrow
  dec r25
one_no_borrow:
  sts ONE_LOW, r24
  sts ONE_HIGH, r25
  movw r26, r24                ; keep the length of a ONE in r27:r26
.endif

  ;mov r19, r27
  ;rcall send_hex
  ;mov r19, r26
  ;rcall send_hex

  ;; Maybe we don't need this.
  movw r24, r4                 ; ZERO = LONG / 4 + 5
  lsr r25
  ror r24
  lsr r25
  ror r24
  add r24, r15 
  brcc zero_no_carry
  inc r25
zero_no_carry:
  sts ZERO_LOW, r24
  sts ZERO_HIGH, r25
  movw r26, r24                ; keep the length of a ONE in r27:r26

  ;; Now time to find the start marker
  rcall wait_for_start_marker

  ;; Now read in all bits and send over rs232
read_next_byte:
  ldi r31, 8
read_next_bit:
  rcall wait_for_black
  rcall count_black

  ;; 16 bit compare to see if r5:r4 >= r25:r24, if so we are done with track
  cp r5, r25
  breq check_high_same_is_done
  brlo not_done_with_track            ;; if r5 < r25, not_done
  brsh done_with_track                ;; if r5 > r25, we are done
check_high_same_is_done:
  cp r4, r24
  brsh done_with_track                ;; if r5 == r25 and r4 >= r24, we are done
not_done_with_track:

  lsl r19

.if 0
  ;; 16 bit compare to see if r5:r4 < r27:r26 (LONG), if so bit is 0
  cp r5, r27
  breq check_high_same_is_done_l1
  brlo bit_is_zero                    ;; if r5 < r27, bit is 0
  brsh bit_is_one                     ;; if r5 > r27, bit is 1
check_high_same_is_done_l1:
  cp r4, r26
  brlo bit_is_zero                    ;; if r5 == r27 and r4 < r26, bit is 0
bit_is_one:
  ori r19, 1
bit_is_zero:
  dec r31
  brne read_next_bit
.endif

  ;; 16 bit compare to see if r5:r4 > r27:r26 (SHORT), if so bit is 1
  cp r5, r27
  breq check_high_same_is_done_l1
  brlo bit_is_zero                     ;; if r5 < r27, bit is 0
  brsh bit_is_one                      ;; if r5 > r27, bit is 1
check_high_same_is_done_l1:
  cp r4, r26
  brlo bit_is_zero                    ;; if r27 == r5 and r4 < r26, bit is 0
bit_is_one:
  ori r19, 1
bit_is_zero:
  dec r31
  brne read_next_bit

  rcall send_char
  ;rcall send_hex
  rjmp read_next_byte

done_with_track:

  ret

;; wait_for_start_marker() { while(black_count < start marker); }
wait_for_start_marker:
  lds r24, LONG_LOW
  lds r25, LONG_HIGH
wait_for_start_marker_repeat:
  rcall wait_for_black
  rcall count_black

  ;; 16 bit compare to see if r25:r24 < r5:r4, if so we found start marker
  cp r5, r25
  breq check_high_same
  brlo wait_for_start_marker_repeat   ;; if r5 < r25, repeat
  brsh wait_for_start_marker_exit     ;; if r5 > r25, we hit the mark
check_high_same:
  cp r4, r24
  brlo wait_for_start_marker_repeat   ;; if r5 == r25 and r4 < r24, repeat
wait_for_start_marker_exit:
  ret

;; wait_for_black() { while(sensor_is_over_white);  return r5:r4 (white count);}
wait_for_black:
  clr r4
  clr r5
wait_for_black_repeat:
  inc r4
  brne repeat_black_wait_nc
  inc r5
repeat_black_wait_nc:
  rcall read_sensor
  tst r22
  breq wait_for_black_repeat
  ret

;; count_black() { while(sensor_is_over_black);  return r5:r4 (black-count); }
count_black:
  clr r4
  clr r5
count_black_repeat:
  inc r4
  brne count_black_nc
  inc r5
count_black_nc:
  rcall read_sensor
  tst r22
  brne count_black_repeat
  ret

.if 0
repeat_read:
  ;in r19, UDR
  ;rcall send_char       ; echo for debug
  rcall read_sensor
  rjmp repeat_read
.endif

; r22: return value 0 for white, 1 for black
; Approx 3 + 3*12 * 5us = 195us.
read_sensor:
  sbi DDRB, 1
  sbi PORTB, 1
  ldi r21, 0
wait_for_charge:
  cpi r21, 3
  brlo wait_for_charge

  cbi PORTB, 1
  cbi DDRB, 1
  ldi r21, 0
wait_for_discharge:
  cpi r21, 3 * 12
  brlo wait_for_discharge

  ;; FIXME - Could move this to PB0 and remove an instruction here
  in r22, PINB
  lsr r22
  andi r22, 1

  sbrs r22, 0
  cbi PORTB, 0
  sbrc r22, 0
  sbi PORTB, 0

  ;sbis PINB, 1
  ;ldi r22, 0
  ;sbic PINB, 1
  ;ldi r22, 1

  ret

; void send_char(r19)  : r14 trashed
send_char:
  in r14, UCSRA       ; check to see if it's okay to send a char
  sbrs r14, UDRE
  rjmp send_char      ; if it's not okay, loop around :)
  out UDR, r19        ; output a char over rs232
  ret

;; TIMER0 8 bit interrupt for sensor
timer0_interrupt:
  in r7, SREG

  inc r21

  out SREG, r7
  reti

;; TIMER1 16 bit interrupt for servos
timer1_interrupt:
  in r7, SREG

  inc r10

  cp r10, r11
  brne servo0_ignore
  cbi PORTB, 2
servo0_ignore:

  cp r10, r12
  brne servo1_ignore
  cbi PORTB, 3
servo1_ignore:

  ;; check if we are at the end of 20ms
  cp r10, r8
  brne exit_interrupt

  mov r10, r0 
  sbi PORTB, 3
  cp r11, r0
  breq exit_interrupt
  ;cpse r11, r0     FIXME - use this instruction instead of the two above
  sbi PORTB, 2

exit_interrupt:
  out SREG, r7
  reti

; void send_nibble(r19)  : r14, r18 trashed
send_nibble:
  cpi r19, 10
  brlo hex_nibble_under_10
  subi r19, 10
  ldi r18, 'A'
  add r19, r18
  rcall send_char
  ret
hex_nibble_under_10:
  ldi r18, '0'
  add r19, r18
  rcall send_char
  ret

; void send_hex(r19)  : r13, r14, r18 trashed
send_hex:
  mov r13, r19
  swap r19
  andi r19, 0x0f
  rcall send_nibble
  mov r19, r13
  andi r19, 0x0f
  rcall send_nibble
  ldi r19, ' '
  rcall send_char
  ret

signature:
.db "PAPER-ROM - Copyright 2013 - Michael Kohn - Version 0.01",0




